from sgmarkets_api_analytics_rates._obj_from_dict import ObjFromDict
from sgmarkets_api_analytics_rates.request_curves_price import RequestCurvesPrices
from sgmarkets_api_analytics_rates.response_curves_price import ResponseCurvesPrice
from sgmarkets_api_analytics_rates.slice_curves_price import SliceCurvesPrice


dic_endpoint = {
    'v1_curves_price': {
        'request': RequestCurvesPrices,
        'response': ResponseCurvesPrice,
        'slice': SliceCurvesPrice
    }
}

endpoint = ObjFromDict(dic_endpoint)

if __name__ == '__main__':
    ep = endpoint
    ep.request()



import numpy as np
import pandas as pd
from itertools import product

ROUND_NB_DIGIT = 2


def get_diff(data, start_date, end_date):
    """
    get difference between 2 dates
    """
    mask_1 = (data.index.get_level_values('date') == end_date)
    mask_2 = (data.index.get_level_values('date') == start_date)
    one = data.loc[mask_1]
    one.index = one.index.droplevel('date')
    two = data.loc[mask_2]
    two.index = two.index.droplevel('date')
    # res=res.index.droplevel('date')
    return one - two


def get_max(data, resize=False, *args):
    """
    get max
    TBC
    """
    if resize:
        data = select_data(data, *args)
    x = data.groupby(level='expiry').apply(max)
    return x.reindex(data.index.get_level_values('expiry').unique(), axis=0)


def get_min(data, resize=False, *args):
    """
    get min
    TBC
    """
    if resize:
        data = select_data(data, *args)
    x = data.groupby(level='expiry').apply(min)
    return x.reindex(data.index.get_level_values('expiry').unique(), axis=0)


def get_surface(data, date):
    """
    get surface for given date
    TBC
    """
    mask = (data.index.get_level_values('date') == date)
    x = data.loc[mask]
    return drop_level(x, 'date')


def select_data(data, args):
    """
    select data from dataframe
    args:
        for discrete values
            field, value1, value2, etc
        for range of values
            field, value_start, value_end, 'range'

    returns subset dataframe
    TBC
    """
    arg = args
    if arg[0] == 'date':
        if len(arg) == 4 and arg[3] == 'range':
            mask = (data.index.get_level_values(arg[0]) >= arg[1]) & (
                data.index.get_level_values(arg[0]) <= arg[2])
        else:
            mask = (data.index.get_level_values(arg[0]) == arg[1])
            for ar in arg[2:len(arg)]:
                mask = mask | (data.index.get_level_values(arg[0]) == ar)
    else:
        mask = (data.index.get_level_values(arg[0]) == arg[1])
        for ar in arg[2:len(arg)]:
            mask = mask | (data.index.get_level_values(arg[0]) == ar)

    return data.loc[mask]


def drop_level(data, index):
    """
    drop level of input dataframe - inplace
    TBC
    """
    data.index = data.index.droplevel(index)
    return data


def get_percentile(data, date, resize=False, *args):
    """
    get percentile rank from input dataframe
    TBC
    """
    if resize:
        data = select_data(data, *args)
    return get_surface(data.groupby(level='expiry').rank(pct=True), date)


def get_mean(data, resize=False, *args):
    """
    get mean
    TBC
    """
    if resize:
        data = select_data(data, *args)
    data = data.dropna()
    x = data.groupby(level='expiry').apply(np.mean)
    return x.reindex(data.index.get_level_values('expiry').unique(), axis=0)


def get_std(data, resize=False, *args):
    """
    get standard deviation
    TBC
    """
    if resize:
        data = select_data(data, *args)
    data = data.dropna()
    x = data.groupby(level='expiry').apply(np.std)
    return x.reindex(data.index.get_level_values('expiry').unique(), axis=0)


def get_z_score(data, end_dt, resize=False, *args):
    """
    get z-score
    TBC
    """
    if resize:
        data = select_data(data, *args)
    return (get_surface(data, end_dt) - get_mean(data)) / get_std(data)


def build_weight(spread, weight, **kwargs):
    """
    build weight
    TBC
    """
    if (len(spread) is not len(weight)):
        if kwargs == {}:
            kwargs.setdefault('type', 'S')
        else:
            kwargs = kwargs['kwargs']

        if (len(spread) % 2 == 0 or (len(spread) % 2 == 0 and len(spread) % 3 == 0)) and (kwargs['type'].upper() != 'FLY' or kwargs['type'].upper() != 'F'):
            print('spread - weight len mismatch by default spread weighted (-1,1)')
            w = np.repeat(1, len(spread))
            pos = np.arange(1, len(w), 2)
            w[pos] = w[pos] * -1
        elif (len(spread) % 3 == 0 or (len(spread) % 2 == 0 and len(spread) % 3 == 0)) and (kwargs['type'].upper() == 'FLY' or kwargs['type'].upper() == 'F'):
            print('fly - weight len mismatch by default fly weighted (-1,2,-1)')
            w = np.repeat(-1, len(spread))
            pos = np.arange(1, len(w), 2)
            w[pos] = w[pos] * -2
    else:
        w = weight
    return w

# def build_spread(data, spread, typ_, resize=False, *args):
#     """
#     build spread
#     TBC
#     """
#     if resize:
#         data = select_data(data, args)
#     res = pd.DataFrame()
#     if typ_.upper() == "S" or typ_.upper() == "SLOPE":
#         for s in spread:
#             res[s[0] + s[1]] = data[s[1]] - data[s[0]]
#         return res
#     if typ_.upper() == "C" or typ._upper() == "CALENDAR":
#         from collections import OrderedDict
#         dic = OrderedDict()
#         for s in spread:
#             dic[(s[0] + s[1])] = data.loc[s[1]] - data.loc[s[0]]
#         res = pd.concat(dic.values(), keys=dic.keys())
#         res.index.names = ('expiry', 'date')
#         return res


def build_spread(data, spread, axis="T", weight=[], resize=False, *args, **kwargs):
    """
    build weight
    TBC
    """
    if resize:
        data = select_data(data, args)
    res = pd.DataFrame()
    w = build_weight(spread[0], weight, **kwargs)
    if axis.upper() == 'T' or axis.upper() == 'TENOR':
        for s in spread:
            nme = str()
            for i, e in enumerate(s):
                if i == 0:
                    res_tmp = data[s[i]] * w[i]
                else:
                    res_tmp = res_tmp + data[s[i]] * w[i]
                nme = nme + s[i]
            res[nme] = res_tmp
        return res

    if axis.upper() == 'E' or axis.upper() == 'EXPIRY':
        from collections import OrderedDict
        dic = OrderedDict()
        idx = data.loc[spread[0][0]].index
        col = data.loc[spread[0][0]].columns
        for s in spread:
            res_tmp = pd.DataFrame(0, index=idx, columns=col)
            nme = str()
            for i, e in enumerate(s):
                res_tmp = res_tmp + data.loc[s[i]] * w[i]
                nme = nme + s[i]
            dic[nme] = res_tmp
        res = pd.concat(dic.values(), keys=dic.keys())
        res.index.names = ('expiry', 'date')
        return res

def swap_prices(rates, dv01, nominal = int(1e6), w=None, hedge='None'):
    if w is None:
        w2 = [1]*len(rates.columns)
    else:
        w2 = []
        for ww in w:
            if isinstance(ww, str):
                if ww.upper() == 'PAYER' or ww.upper() == 'P':
                    w2.append(1)
                elif ww.upper() == 'RECEIVER' or ww.upper() == 'R':
                    w2.append(-1)
            else:
                w2.append(ww)

    if hedge.upper() == 'DV01':

        w_ = [1/(dv01.iloc[0,i]/dv01.iloc[0, 0]) for i in range(len(dv01.columns))]
        w2 = [a*b for a,b in zip(w_, w2)]

    ret = rates.diff(1).fillna(0)
    prices = ret * dv01 * w2 * nominal
    prices.columns = [str(w[c]) +' '+ list(prices.columns)[c][0]+' '+list(prices.columns)[c][1] for c in range(len(prices.columns))]
    portfolio_price = prices.sum(axis=1)
    res = pd.concat([prices, portfolio_price], axis = 1)
    res.columns = list(prices.columns) + ['strategy price']
    for c in prices.columns:
        res[c + ' PnL'] = prices[c].cumsum()
    res['strategy PnL'] = res['strategy price'].cumsum()
    return w2, res


def build_rolldown(couples, period):
    cpl = []
    ten = []
    exp = []
    if 'y' in period:
        period = int(period.replace('y', '')) * 12
    else:
        period = int(period.replace('m', ''))
    for l in couples:
        exp.append(l[0])
        ten.append(l[1])
        if 'y' in l[0]:
            e = int(l[0].replace('y', '')) * 12
        else:
            e = int(l[0].replace('m', ''))

        if 'y' in l[1]:
            t = int(l[1].replace('y', '')) * 12
        else:
            t = int(l[1].replace('m', ''))

        diff = e - period
        if diff >= 0:
            if diff % 12 == 0 and diff != 0:
                e = str(int(diff / 12)) + 'y'
            else:
                e = str(int(diff)) + 'm'
            if t % 12 == 0:
                t = str(int(t / 12)) + 'y'
            else:
                t = str(int(t)) + 'm'
            exp.append(e)
            ten.append(t)
            cpl.append(((e, t), l))

        else:
            exp.append('0m')
            diff_t = t + diff
            if diff_t % 12 == 0:
                t = str(int(diff_t / 12)) + 'y'
            else:
                t = str(int(diff_t)) + 'm'
            ten.append(t)
            cpl.append((('0m', t), l))

    return ','.join(list(set(exp))), ','.join(list(set(ten))), cpl


def build_couples(exp,ten):
    return [(e,t) for e, t in product(exp,ten)]

def rolldown(df, cpl, w=None):
        if not w:
            w = [-1, 1]
        res = {}
        for c in cpl:
            d = w[0] * df[c[0][1]].loc[c[0][0]] + w[1] * df[c[1][1]].loc[c[1][0]]
            d = pd.DataFrame(d)
            d.columns = [c[1][1]]
            if not c[1][0] in res.keys():
                res[c[1][0]] = d
            else:
                res[c[1][0]] = pd.concat([res[c[1][0]], d], axis=1)
        return pd.concat(res)


def spread_cross(df, cpl, w=None):
    if not w:
        w = [-1, 1]
    res = pd.DataFrame()
    for c in cpl:
        res[c[1][0] + c[1][1] + '-' + c[0][0] + c[0][1]] = w[0] * df[c[0][1]].loc[c[0][0]] + w[1] * df[c[1][1]].loc[
            c[1][0]]
    return res


def build_tenor_rolldown_spot(list_of_tenor, period):
    if not isinstance(list_of_tenor, list):
        list_of_tenor = list_of_tenor.split(',')
    ten = []
    for l in list_of_tenor:
        if 'y' in l:
            n = int(l.replace('y', '')) * 12 - period
            assert n >= 0, 'period should be inferior or equal to the shortest expiries'
            ten.append(str(n) + 'm')
        else:
            n = int(l.replace('m', '')) - period
            assert n >= 0, 'period should be inferior or equal to the shortest expiries'
            ten.append(str(n) + 'm')
    new_ten = list_of_tenor + ten

    return ','.join(new_ten), [(ten[i], list_of_tenor[i]) for i in range(len(ten))]


def build_carry(couples, period):
    cpl = []
    ten = []
    exp = []
    if 'y' in period:
        period = int(period.replace('y', '')) * 12
    else:
        period = int(period.replace('m', ''))
    for l in couples:
        exp.append(l[0])
        ten.append(l[1])
        if 'y' in l[0]:
            e = int(l[0].replace('y', '')) * 12
        else:
            e = int(l[0].replace('m', ''))

        if 'y' in l[1]:
            t = int(l[1].replace('y', '')) * 12
        else:
            t = int(l[1].replace('m', ''))

        diff_exp = period - e
        if diff_exp <= 0:
            exp.append(l[0])
            ten.append(l[1])
            e_ = l[0]
            t = l[1]
            l_2 = l

        else:
            e = period - e
            if e % 12 == 0:
                e_ = str(int(e / 12)) + 'y'
            else:
                e_ = str(int(e)) + 'm'
            exp.append(e_)
            t = t - e
            if t % 12 == 0:
                t = str(int(t / 12)) + 'y'
            else:
                t = str(int(t)) + 'm'
            ten.append(t)
            l_2 = l
            l = ('0m', l[1])
        cpl.append(((e_, t), l, l_2))

    return ','.join(list(set(exp))), ','.join(list(set(ten))), cpl


def carry_approximation(df, cpl, w=None):
    if not w:
        w = [1, -1]
    res = {}
    for c in cpl:
        d = w[0] * df[c[0][1]].loc[c[0][0]] + w[1] * df[c[1][1]].loc[c[1][0]]
        d = pd.DataFrame(d)
        d.columns = [c[2][1]]
        if not c[2][0] in res.keys():
            res[c[2][0]] = d
        else:
            res[c[2][0]] = pd.concat([res[c[2][0]], d], axis=1)
    return pd.concat(res)


import copy
import operator

import itertools as it
import pandas as pd
import datetime as dt

from IPython.display import display, Markdown
from timeit import default_timer as timer
from dateutil.relativedelta import relativedelta

from sgmarkets_api_analytics_rates.response_curves_price import ResponseCurvesPrice
from sgmarkets_api_analytics_rates._util import Util


CHUNK_INCREMENT = relativedelta(months=1)


class RequestCurvesPrices(object):

    #TODO: put start / end date parameters in __init__ and perform isinstance checks
    def __init__(self, chunk_increment=CHUNK_INCREMENT):

        dic_url = {'base_url': 'https://analytics-api.sgmarkets.com',
                   'service': '/rates',
                   'endpoint': '/v1/curves/price'}
        self.url = Util.build_url(dic_url)
        self.chunk_increment = None #chunk_increment
        self.dic_api = None
        self.li_dic_api = None
        self.df_top = None
        self.df_leg = None
        self.all_combination = True
        self.outputFormat = 'GroupByDate'

        self.leg_keys = [
            'ticker',
            'expiry',
            'tenor',
            'type'
        ]

    def _to_list(self, val):
        """
        """
        if isinstance(val, int):
            return [val]
        elif ',' not in val:
            return [val]
        else:
            li_val = val.split(',')
            li_val = [e.strip() for e in li_val]
            return li_val

    def info(self):
        pass

    def expand(self):

        dic_in = copy.copy(self.__dict__)
        keys = ['dic_api',
                'leg_keys',
                'df_top',
                'df_leg',
                'li_dic_api',
                'url',
                'chunk_increment']

        for key in keys:
            dic_in.pop(key)

        # change keys to format to API keys
        change_key = {
            'start_date': 'startDate',
            'end_date': 'endDate',
            'curve': 'ticker'
        }

        for key in change_key:
            dic_in[change_key[key]] = dic_in[key]
            dic_in.pop(key)

        # build dic as a prelude to dic_api
        self.dic_input = {}
        for k, v in dic_in.items():
            if k in self.leg_keys:
                # make params in leg to list
                self.dic_input[k] = self._to_list(v)
            else:
                # top params unchanged
                self.dic_input[k] = v

        # cartesian product of all params in leg
        assert isinstance(self.all_combination,bool), """ all_combination should be a boolean """


        if self.all_combination:
            items = [(k, v)
                     for k, v in self.dic_input.items() if k in self.leg_keys]
            keys = [e[0] for e in items]
            values = [e[1] for e in items]
            li_leg = [dict(zip(keys, e)) for e in it.product(*values)]
        else:
            max_ = self.check_parameters()
            li_leg = []
            for i in range(max_):
                dic = {k: v[i] for k, v in self.dic_input.items()
                       if k in self.leg_keys and v[i] is not "None"}
                li_leg.append(dic)

        # modify li_leg values to include 'term' key from (expiry, tenor) pairs
        # TODO: remove for loop and construct term key earlier in the expand() method ?
        for e in li_leg:
            e['term'] = '{}x{}'.format(e['expiry'], e['tenor'])
            e.pop('expiry')
            e.pop('tenor')

        # merge top params and params in legs expanded by cartesian product
        self.dic_api = {k: v
                        for k, v in self.dic_input.items() if not k in self.leg_keys}
        self.dic_api['curves'] = li_leg
        # building term key from expiry tenor couple
        dic_top = {k: str(v)
                   for k, v in self.dic_api.items() if k != 'curves'}
        self.df_top = pd.DataFrame(dic_top, index=['Value']).T

        self.df_leg = pd.DataFrame(li_leg)
        self.li_dic_api = self._split_input_by_dates(li_leg)

    def check_parameters(self):
        """
        check and complete the input_dic if necessary
        """

        len_ = {k: len(v) for k, v in self.dic_input.items()
                if isinstance(v, list)}
        max_ = max(len_.items(), key=operator.itemgetter(1))[1]
        err = False
        for k, v in len_.items():
            if v is not max_:

                msg = '**<span style="color:red;">WARNING</span> -**' \
                          + 'improper combinaison of parameters resulting in an unclear strategy' \
                          + ', remainder of the division of the longest parameters and others should be 0 '
                assert (max_ % v) is 0, display(Markdown(msg))
                err = True
                self.dic_input[k] = self.dic_input[k] * int(max_ / v)

        if err:
            msg = '**<span style="color:red;">WARNING</span> -**' \
                  + 'improper combinaison of parameters resulting in an unclear strategy' \
                  + ', by default missing parameters will be filled by repetition'
            display(Markdown(msg))

        return max_

    def _split_input_by_dates(self, li_leg):
        start_date = dt.datetime.strptime(self.dic_api['startDate'], '%Y-%m-%d')
        end_date = dt.datetime.strptime(self.dic_api['endDate'], '%Y-%m-%d')

        li_date_chunk = self._chunk_dates(start_date, end_date)
        li_dic_api = []
        for start, end in li_date_chunk:
            dic_api_chunk = {k: v
                             for k, v in self.dic_input.items() if not k in self.leg_keys}
            dic_api_chunk['startDate'] = start
            dic_api_chunk['endDate'] = end
            li_leg_chunk = copy.deepcopy(li_leg)
            dic_api_chunk['curves'] = li_leg_chunk
            li_dic_api.append(dic_api_chunk)

        return li_dic_api

    #TODO: rewrite method to avoir while loop
    def _chunk_dates(self, start_date, end_date):
        nb_days = end_date - start_date
        max_nb_of_days = 2000/(len(self.dic_api['curves']))
        nb_of_chunks = max(1, int(nb_days.days/max_nb_of_days)+1)
        self.chunk_increment = relativedelta(days=(nb_days.days/nb_of_chunks))
        fmt = '%Y-%m-%d'
        _start = start_date
        _end = end_date
        chunks = []
        t0 = start_date
        t1 = min(_start + self.chunk_increment, _end)
        if t1 == _end:
            chunks.append((dt.datetime.strftime(t0,fmt), dt.datetime.strftime(t1,fmt)))
            return chunks

        while t1 <= _end:
            chunks.append((t0, t1))
            _end_chunk = t1
            t0 = t1 + relativedelta(days=1)
            t1 = _end_chunk + self.chunk_increment

        #_last_chunk = chunks[-1][-1]
        #chunks.append((_last_chunk + relativedelta(days=1), end_date))
        chunks_str = [(dt.datetime.strftime(start, fmt), dt.datetime.strftime(end, fmt)) for start, end in chunks]
        return chunks_str

    def call_api(self, api, debug=False):
        t0 = timer()
        print('calling API...')
        li_raw_response = []
        for k, dic_api in enumerate(self.li_dic_api):
            raw_response = api.post(self.url, payload=dic_api)
            if debug:
                print('chunk {} raw response\n{}'.format(k, raw_response))
            li_raw_response.append(raw_response)
            t1 = timer()
            print('chunk {}/{} done in {:.2f} s'.format(1 + k,
                                                        len(self.li_dic_api),
                                                        t1 - t0))
        if debug:
            print('*** START DEBUG ***\n{}\n*** END DEBUG ***'.format(li_raw_response))

        response = ResponseCurvesPrice(li_raw_data=li_raw_response,
                                 obj_req=self)


        return response






import pandas as pd

from collections import OrderedDict


class Util:

    @staticmethod
    def get_unique_list(my_list):
        """
        """
        unique_list = list(OrderedDict((e, None) for e in my_list))
        unique_list = [e for e in unique_list if pd.notnull(e)]
        return unique_list

    @staticmethod
    def build_url(dic_url):
        return '{}{}{}'.format(dic_url['base_url'], dic_url['service'], dic_url['endpoint'])

    @staticmethod
    def date_to_str(ts):
        """
        """
        return ts.strftime('%Y-%m-%d')
